import json
import select
import socket
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Collection, Iterable, List, Optional, Tuple, Union

from platypush.config import Config
from platypush.context import get_bus
from platypush.message.event.music.snapcast import (
    ClientVolumeChangeEvent,
    GroupMuteChangeEvent,
    ClientConnectedEvent,
    ClientDisconnectedEvent,
    ClientLatencyChangeEvent,
    ClientNameChangeEvent,
    GroupStreamChangeEvent,
    StreamUpdateEvent,
    ServerUpdateEvent,
)
from platypush.plugins import RunnablePlugin, action


class MusicSnapcastPlugin(RunnablePlugin):
    """
    Plugin to interact with a `Snapcast <https://github.com/badaix/snapcast>`_
    instance, control clients mute status, volume, playback etc.

    See https://github.com/badaix/snapcast/blob/master/doc/json_rpc_api/v2_0_0.md
    for further reference about the returned API types.
    """

    _DEFAULT_SNAPCAST_PORT = 1705
    _DEFAULT_POLL_SECONDS = 10  # Poll servers each 10 seconds
    _SOCKET_EOL = '\r\n'.encode()

    def __init__(
        self,
        host: Optional[str] = None,
        port: int = _DEFAULT_SNAPCAST_PORT,
        hosts: Optional[Iterable[dict]] = None,
        poll_interval: Optional[float] = _DEFAULT_POLL_SECONDS,
        **kwargs,
    ):
        """
        :param host: Default Snapcast server host.
        :param port: Default Snapcast server control port (default: 1705).
        :param hosts: If specified, then the provided list of Snapcast servers
            will be monitored, rather than just the one provided on ``host``.
            This setting can be used either in conjunction with ``host`` (in
            that case, if the ``host`` is not specified on a request then
            ``host`` will be used as a fallback), or on its own (in that case
            requests with no host specified will target the first server in the
            list). Note however that either ``host`` or ``hosts`` must be
            provided. Format:

                .. code-block:: yaml

                    hosts:
                        - host: localhost
                          port: 1705  # Default port

                        - host: snapcast.example.com
                          port: 9999

        :param poll_seconds: How often the plugin will poll remote servers for
            status updates (default: 10 seconds).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)

        self._hosts = self._get_hosts(host=host, port=port, hosts=hosts)
        assert self._hosts, 'No Snapcast hosts specified'
        self._latest_req_id = 0
        self._latest_req_id_lock = threading.RLock()

        self._socks = {}
        self._threads = {}
        self._statuses = {}

    @property
    def host(self) -> str:
        return self._hosts[0][0]

    @property
    def port(self) -> int:
        if not getattr(self, '_hosts', None):
            return self._DEFAULT_SNAPCAST_PORT

        return self._hosts[0][1]

    def _get_hosts(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        hosts: Optional[Iterable[dict]] = None,
    ) -> List[Tuple[str, int]]:
        ret = []
        if hosts:
            assert all(
                isinstance(h, dict) and h.get('host') for h in hosts
            ), f'Expected a list of dicts with host and port keys, got: {hosts}'

            ret.extend((h['host'], h.get('port', self.port)) for h in hosts)

        if host and port:
            ret.insert(0, (host, port))

        return list(dict.fromkeys(ret))

    def _next_req_id(self):
        with self._latest_req_id_lock:
            self._latest_req_id += 1
            return self._latest_req_id

    def _connect(self, host: str, port: int, reuse: bool = False):
        if reuse and self._socks.get(host) and self._socks[host].fileno() >= 0:
            return self._socks[host]

        self.logger.debug('Connecting to %s:%d', host, port)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((host, port))

        if reuse:
            self._socks[host] = sock

        self.logger.info('Connected to %s:%d', host, port)
        return sock

    def _disconnect(self, host: str, port: int):
        sock = self._socks.get(host)
        if not sock:
            self.logger.debug('Not connected to %s:%d', host, port)
            return

        try:
            sock.close()
        except Exception as e:
            self.logger.warning(
                'Exception while disconnecting from %s:%d: %s', host, port, e
            )
        finally:
            self._socks[host] = None

    @classmethod
    def _send(cls, sock: socket.socket, req: Union[dict, str, bytes]):
        if isinstance(req, dict):
            req = json.dumps(req)
        if isinstance(req, str):
            req = req.encode()

        assert isinstance(
            req, bytes
        ), f'Unsupported type {type(req)} for Snapcast request'
        sock.send(req + cls._SOCKET_EOL)

    @classmethod
    def _recv_result(cls, sock: socket.socket):
        msg = cls._recv(sock)
        if not msg:
            return None

        return msg.get('result')

    @classmethod
    def _recv(cls, sock: socket.socket):
        sock.setblocking(False)
        buf = b''

        while buf[-2:] != cls._SOCKET_EOL:
            ready = select.select([sock], [], [], 0.5)
            if ready[0]:
                ch = sock.recv(1)
                if not ch:
                    raise ConnectionError('Connection reset by peer')

                buf += ch
            else:
                return None

        return json.loads(buf)

    @classmethod
    def _parse_event(cls, host, msg):
        evt = None

        if msg.get('method') == 'Client.OnVolumeChanged':
            client_id = msg.get('params', {}).get('id')
            volume = msg.get('params', {}).get('volume', {}).get('percent')
            muted = msg.get('params', {}).get('volume', {}).get('muted')
            evt = ClientVolumeChangeEvent(
                host=host, client=client_id, volume=volume, muted=muted
            )
        elif msg.get('method') == 'Group.OnMute':
            group_id = msg.get('params', {}).get('id')
            muted = msg.get('params', {}).get('mute')
            evt = GroupMuteChangeEvent(host=host, group=group_id, muted=muted)
        elif msg.get('method') == 'Client.OnConnect':
            client = msg.get('params', {}).get('client')
            evt = ClientConnectedEvent(host=host, client=client)
        elif msg.get('method') == 'Client.OnDisconnect':
            client = msg.get('params', {}).get('client')
            evt = ClientDisconnectedEvent(host=host, client=client)
        elif msg.get('method') == 'Client.OnLatencyChanged':
            client = msg.get('params', {}).get('id')
            latency = msg.get('params', {}).get('latency')
            evt = ClientLatencyChangeEvent(host=host, client=client, latency=latency)
        elif msg.get('method') == 'Client.OnNameChanged':
            client = msg.get('params', {}).get('id')
            name = msg.get('params', {}).get('name')
            evt = ClientNameChangeEvent(host=host, client=client, name=name)
        elif msg.get('method') == 'Group.OnStreamChanged':
            group_id = msg.get('params', {}).get('id')
            stream_id = msg.get('params', {}).get('stream_id')
            evt = GroupStreamChangeEvent(host=host, group=group_id, stream=stream_id)
        elif msg.get('method') == 'Stream.OnUpdate':
            stream_id = msg.get('params', {}).get('stream_id')
            stream = msg.get('params', {}).get('stream')
            evt = StreamUpdateEvent(host=host, stream_id=stream_id, stream=stream)
        elif msg.get('method') == 'Server.OnUpdate':
            server = msg.get('params', {}).get('server')
            evt = ServerUpdateEvent(host=host, server=server)

        return evt

    def _event_listener(self, host: str, port: int):
        def _thread():
            while not self.should_stop():
                try:
                    sock = self._connect(host, port, reuse=True)
                    msgs = self._recv(sock)

                    if msgs is None:
                        continue
                    if not isinstance(msgs, list):
                        msgs = [msgs]

                    for msg in msgs:
                        self.logger.debug(
                            'Received message on %s:%d: %s', host, port, msg
                        )

                        evt = self._parse_event(host=host, msg=msg)
                        if evt:
                            get_bus().post(evt)
                except Exception as e:
                    self.logger.warning(
                        'Exception while getting the status of the Snapcast server %s:%d: %s',
                        host,
                        port,
                        e,
                    )

                    self._disconnect(host, port)
                finally:
                    if self.poll_interval:
                        time.sleep(self.poll_interval)

        return _thread

    def _get_group(self, sock: socket.socket, group: str):
        for g in self._get_status(sock).get('groups', []):
            if group == g.get('id') or group == g.get('name'):
                return g

        return None

    def _get_client(self, sock: socket.socket, client: str):
        for g in self._get_status(sock).get('groups', []):
            clients = g.get('clients', [])

            for c in clients:
                if (
                    client == c.get('id')
                    or client == c.get('name')
                    or client == c.get('host', {}).get('name')
                    or client == c.get('host', {}).get('ip')
                ):
                    c['group_id'] = g.get('id')
                    return c

        return None

    def _get_status(self, sock: socket.socket) -> dict:
        request = {
            'id': self._next_req_id(),
            'jsonrpc': '2.0',
            'method': 'Server.GetStatus',
        }

        self._send(sock, request)
        return (self._recv_result(sock) or {}).get('server', {})

    def _status(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        client: Optional[str] = None,
        group: Optional[str] = None,
    ):
        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            if client:
                return self._get_client(sock, client)
            if group:
                return self._get_group(sock, group)

            return self._get_status(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    def _status_response(
        self,
        host: str,
        port: int,
        client: Optional[str] = None,
        group: Optional[str] = None,
    ):
        sock = None
        try:
            sock = self._connect(host, port)
            if client:
                return self._get_client(sock, client)
            if group:
                return self._get_group(sock, group)

            return {'host': host, 'port': port, **self._get_status(sock)}
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def status(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        client: Optional[str] = None,
        group: Optional[str] = None,
    ):
        """
        Get the current status of a Snapcast server, client or group.

        If not host, client or group is specified, the action will return the
        status of all the Snapcast servers.

        :param host: Snapcast server to query (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        :param client: Client ID or name (default: None)
        :param group: Group ID or name (default: None)

        :returns: dict. Example:

            .. code-block:: json

                "output": {
                    "host": "localhost",
                    "port": 1705,
                    "groups": [
                        {
                        "clients": [
                            {
                            "config": {
                                "instance": 1,
                                "latency": 0,
                                "name": "",
                                "volume": {
                                "muted": false,
                                "percent": 96
                                }
                            },
                            "connected": true,
                            "host": {
                                "arch": "x86_64",
                                "ip": "YOUR_IP",
                                "mac": "YOUR_MAC",
                                "name": "YOUR_NAME",
                                "os": "YOUR_OS"
                            },
                            "id": "YOUR_ID",
                            "lastSeen": {
                                "sec": 1546648311,
                                "usec": 86011
                            },
                            "snapclient": {
                                "name": "Snapclient",
                                "protocolVersion": 2,
                                "version": "0.15.0"
                            }
                            }
                        ],
                        "id": "YOUR_ID",
                        "muted": false,
                        "name": "",
                        "stream_id": "mopidy"
                        }
                    ],
                    "server": {
                        "host": {
                        "arch": "armv7l",
                        "ip": "",
                        "mac": "",
                        "name": "YOUR_NAME",
                        "os": "YOUR_OS"
                        },
                        "snapserver": {
                        "controlProtocolVersion": 1,
                        "name": "Snapserver",
                        "protocolVersion": 1,
                        "version": "0.15.0"
                        }
                    },
                    "streams": [
                        {
                            "id": "mopidy",
                            "meta": {
                                "STREAM": "mopidy"
                            },
                            "status": "playing",
                            "uri": {
                                "fragment": "",
                                "host": "",
                                "path": "/tmp/snapfifo",
                                "query": {
                                "buffer_ms": "20",
                                "codec": "pcm",
                                "name": "mopidy",
                                "sampleformat": "48000:16:2"
                                },
                                "raw": "pipe:////tmp/fifo?buffer_ms=20&codec=pcm&name=mopidy&sampleformat=48000:16:2",
                                "scheme": "pipe"
                            }
                        }
                    ]
                }

        """

        if client or group or host:
            return self._status_response(
                host=host or self.host,
                port=port or self.port,
                client=client,
                group=group,
            )

        # Run status in parallel on all the hosts and return a list with all the
        # results
        with ThreadPoolExecutor(max_workers=len(self._hosts)) as executor:
            return list(
                executor.map(lambda h: self._status_response(h[0], h[1]), self._hosts)
            )

    @action
    def mute(
        self,
        client: Optional[str] = None,
        group: Optional[str] = None,
        mute: Optional[bool] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Set the mute status of a connected client or group

        :param client: Client name or ID to mute
        :param group: Group ID to mute
        :param mute: Mute status. If not set, the mute status of the selected
            client/group will be toggled.
        :param host: Snapcast server to query (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        assert client or group, 'Please specify either a client or a group'
        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Group.SetMute' if group else 'Client.SetVolume',
                'params': {},
            }

            if group:
                g = self._get_group(sock, group)
                assert g, f'No such group: {group}'
                cur_muted = g['muted']
                request['params']['id'] = g['id']
                request['params']['mute'] = not cur_muted if mute is None else mute
            elif client:
                c = self._get_client(sock, client)
                assert c, f'No such client: {client}'
                cur_muted = c['config']['volume']['muted']
                request['params']['id'] = c['id']
                request['params']['volume'] = {}
                request['params']['volume']['percent'] = c['config']['volume'][
                    'percent'
                ]
                request['params']['volume']['muted'] = (
                    not cur_muted if mute is None else mute
                )

            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def volume(
        self,
        client: str,
        volume: Optional[int] = None,
        delta: Optional[int] = None,
        mute: Optional[bool] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Set the volume of a connected client.

        :param client: Client name or ID
        :param volume: Absolute volume to set between 0 and 100
        :param delta: Relative volume change in percentage (e.g. +10 or -10)
        :param mute: Set to true or false if you want to toggle the muted status
        :param host: Snapcast server (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        assert not (volume is None and delta is None and mute is None), (
            'Please specify either an absolute volume, a relative delta or '
            + 'a mute status'
        )

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Client.SetVolume',
                'params': {},
            }

            c = self._get_client(sock, client)
            assert c, f'No such client: {client}'
            cur_volume = int(c['config']['volume']['percent'])
            cur_mute = bool(c['config']['volume']['muted'])

            if volume is not None:
                volume = int(volume)
            elif delta is not None:
                volume = cur_volume + int(delta)

            volume = max(0, min(100, volume)) if volume is not None else cur_volume
            if mute is None:
                mute = cur_mute

            request['params']['id'] = c['id']
            request['params']['volume'] = {}
            request['params']['volume']['percent'] = volume
            request['params']['volume']['muted'] = mute
            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def set_client_name(
        self,
        client: str,
        name: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Set/change the name of a connected client

        :param client: Current client name or ID to rename
        :param name: New name
        :param host: Snapcast server (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Client.SetName',
                'params': {},
            }

            c = self._get_client(sock, client)
            assert c, f'No such client: {client}'
            request['params']['id'] = c['id']
            request['params']['name'] = name
            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def set_group_name(
        self,
        group: str,
        name: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Set/change the name of a group

        :param group: Group ID to rename
        :param name: New name
        :param host: Snapcast server (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Group.SetName',
                'params': {
                    'id': group,
                    'name': name,
                },
            }

            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def set_latency(
        self,
        client: str,
        latency: float,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Set/change the latency of a connected client

        :param client: Client name or ID
        :param latency: New latency in milliseconds
        :param host: Snapcast server (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Client.SetLatency',
                'params': {'latency': latency},
            }

            c = self._get_client(sock, client)
            assert c, f'No such client: {client}'
            request['params']['id'] = c['id']
            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def delete_client(
        self, client: str, host: Optional[str] = None, port: Optional[int] = None
    ):
        """
        Delete a client from the Snapcast server

        :param client: Client name or ID
        :param host: Snapcast server (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Server.DeleteClient',
                'params': {},
            }

            c = self._get_client(sock, client)
            assert c, f'No such client: {client}'
            request['params']['id'] = c['id']
            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def group_set_clients(
        self,
        group: str,
        clients: Collection[str],
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Sets the clients for a group on a Snapcast server

        :param group: Group name or ID
        :param clients: List of client names or IDs
        :param host: Snapcast server (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            g = self._get_group(sock, group)
            assert g, f'No such group: {group}'
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Group.SetClients',
                'params': {'id': g['id'], 'clients': []},
            }

            for client in clients:
                c = self._get_client(sock, client)
                assert c, f'No such client: {client}'
                request['params']['clients'].append(c['id'])

            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def group_set_stream(
        self,
        group: str,
        stream_id: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Sets the active stream for a group.

        :param group: Group name or ID
        :param stream_id: Stream ID
        :param host: Snapcast server (default: default configured host)
        :param port: Snapcast server port (default: default configured port)
        """

        sock = None

        try:
            sock = self._connect(host or self.host, port or self.port)
            g = self._get_group(sock, group)
            assert g, f'No such group: {group}'
            request = {
                'id': self._next_req_id(),
                'jsonrpc': '2.0',
                'method': 'Group.SetStream',
                'params': {
                    'id': g['id'],
                    'stream_id': stream_id,
                },
            }

            self._send(sock, request)
            return self._recv_result(sock)
        finally:
            try:
                if sock:
                    sock.close()
            except Exception as e:
                self.logger.warning('Error on socket close: %s', e)

    @action
    def get_playing_streams(self, exclude_local: bool = False):
        """
        Returns the configured remote streams that are currently active and
        unmuted.

        .. warning:: This action is deprecated and mostly kept for
            backward-compatibility purposes, as it doesn't allow the case where
            multiple Snapcast instances can be running on the same host, nor it
            provides additional information other that something is playing on a
            certain host and port. Use :meth:`.status` instead.

        :param exclude_local: Exclude localhost connections (default: False)
        :returns: dict with the host->port mapping. Example:

            .. code-block:: json

                {
                    "hosts": {
                        "server_1": 1705,
                        "server_2": 1705,
                        "server_3": 1705
                    }
                }

        """

        playing_hosts = {}

        def _worker(host: str, port: int):
            try:
                if exclude_local and (
                    host == 'localhost' or host == Config.get('device_id')
                ):
                    return

                server_status = self._status(host=host, port=port) or {}
                device_id: str = Config.get('device_id')  # type: ignore
                client_status = (
                    self._status(host=host, port=port, client=device_id) or {}
                )

                if client_status.get('config', {}).get('volume', {}).get('muted'):
                    return

                group = next(
                    iter(
                        g
                        for g in server_status.get('groups', {})
                        if g.get('id') == client_status.get('group_id')
                    )
                )

                if group.get('muted'):
                    return

                stream = next(
                    iter(
                        s
                        for s in server_status.get('streams', {})
                        if s.get('id') == group.get('stream_id')
                    )
                )

                if stream.get('status') != 'playing':
                    return

                playing_hosts[host] = port
            except Exception as e:
                self.logger.warning(
                    'Error while retrieving the status of Snapcast host at %s:%d: %s',
                    host,
                    port,
                    e,
                )

        workers = []

        for host, port in self._hosts:
            w = threading.Thread(target=_worker, args=(host, port))
            w.start()
            workers.append(w)

        while workers:
            w = workers.pop()
            w.join()

        return {'hosts': playing_hosts}

    def main(self):
        while not self.should_stop():
            for host, port in self._hosts:
                thread_name = f'Snapcast-{host}-{port}'

                self._threads[host] = threading.Thread(
                    target=self._event_listener(host, port), name=thread_name
                )

                self._threads[host].start()

            for thread in self._threads.values():
                thread.join()

            self._threads = {}

    def stop(self):
        for host, sock in self._socks.items():
            if sock:
                try:
                    sock.close()
                except Exception as e:
                    self.logger.warning(
                        'Could not close Snapcast connection to %s: %s: %s',
                        host,
                        type(e),
                        e,
                    )

        super().stop()


# vim:sw=4:ts=4:et:



from concurrent.futures import ProcessPoolExecutor
from logging import getLogger
from typing import Dict, List, Optional

import bluetooth

from platypush.entities.bluetooth import BluetoothService

from .._model import BluetoothServicesBuilder


# pylint: disable=too-few-public-methods
class ServiceDiscoverer:
    """
    Runs the service discovery processes in a pool.
    """

    def __init__(self, max_workers: int = 4, timeout: float = 30):
        self._max_workers = max_workers
        self._timeout = timeout
        self.logger = getLogger(__name__)

    def _discover(self, address: str) -> List[BluetoothService]:
        """
        Inner implementation of the service discovery for a specific device.
        """
        self.logger.info("Discovering services for %s...", address)
        try:
            return BluetoothServicesBuilder.build(
                bluetooth.find_service(address=address)
            )
        except Exception as e:
            self.logger.warning(
                "Failed to discover services for the device %s: %s", address, e
            )
            return []
        finally:
            self.logger.info("Service discovery for %s completed", address)

    def discover(
        self, *addresses: str, timeout: Optional[float] = None
    ) -> Dict[str, List[BluetoothService]]:
        """
        Discover the services for the given addresses. Discovery will happen in
        parallel through a process pool.

        :param addresses: The addresses to scan.
        :param timeout: The timeout in seconds.
        :return: An ``{address: [services]}`` dictionary with the discovered
            services per device.
        """
        discovered_services: Dict[str, List[BluetoothService]] = {}
        with ProcessPoolExecutor(max_workers=self._max_workers) as executor:
            try:
                for i, services in enumerate(
                    executor.map(
                        self._discover, addresses, timeout=timeout or self._timeout
                    )
                ):
                    discovered_services[addresses[i]] = services
            except TimeoutError:
                self.logger.warning("Service discovery timed out.")

        return discovered_services



from dataclasses import dataclass
from typing import List, Optional

from ._battery import Battery
from ._cpu import Cpu
from ._disk import Disk
from ._fan import Fan
from ._memory import MemoryStats, SwapStats
from ._network import NetworkInterface
from ._temperature import Temperature


@dataclass
class SystemInfo:
    """
    Aggregate system info dataclass.
    """

    cpu: Cpu
    memory: MemoryStats
    swap: SwapStats
    disks: List[Disk]
    network: List[NetworkInterface]
    temperature: List[Temperature]
    fans: List[Fan]
    battery: Optional[Battery]



from typing import Dict

from platypush.backend.http.media.handlers import MediaHandler


# Size for the bytes chunk sent over the media streaming infra
STREAMING_CHUNK_SIZE = 4096

# Maximum range size to be sent through the media streamer if Range header
# is not set
STREAMING_BLOCK_SIZE = 3145728

# Name of the Redis variable used to store the media map across several
# Web processes
MEDIA_MAP_VAR = 'platypush__stream_media_map'

MediaMap = Dict[str, MediaHandler]



from abc import ABC, abstractmethod
from http.client import responses
import json
from logging import getLogger
from typing import Optional

from tornado.web import RequestHandler, stream_request_body

from platypush.backend.http.app.utils import logger
from platypush.backend.http.app.utils.auth import AuthStatus, get_auth_status

from ..mixins import PubSubMixin


@stream_request_body
class StreamingRoute(RequestHandler, PubSubMixin, ABC):
    """
    Base class for Tornado streaming routes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = getLogger(__name__)

    def prepare(self):
        """
        Request preparation logic. It performs user authentication if
        ``auth_required`` returns True, and it can be extended/overridden.
        """
        if self.auth_required:
            auth_status = get_auth_status(self.request)
            if auth_status != AuthStatus.OK:
                self.send_error(auth_status.value.code, error=auth_status.value.message)
                return

        self.logger.info(
            'Client %s connected to %s', self.request.remote_ip, self.request.path
        )

    def write_error(self, status_code: int, error: Optional[str] = None, **_):
        """
        Make sure that errors are always returned in JSON format.
        """
        self.set_header("Content-Type", "application/json")
        self.set_status(status_code)
        self.finish(
            json.dumps(
                {"status": status_code, "error": error or responses.get(status_code)}
            )
        )

    @classmethod
    @abstractmethod
    def path(cls) -> str:
        """
        Path/URL pattern for this route.
        """
        raise NotImplementedError()

    @property
    def auth_required(self) -> bool:
        """
        If set to True (default) then this route will require user
        authentication and return 401 if authentication fails.
        """
        return True

    @classmethod
    def _get_redis_queue(cls, *_, **__) -> Optional[str]:
        """
        Returns the Redis channel associated with a given set of arguments.

        This is None by default, and it should be implemented by subclasses if
        required.
        """
        return None

    def forward_stream(self, *args, **kwargs):
        """
        Utility method that does the following:

            1. It listens for new messages on the subscribed Redis channels;
            2. It applies a filter on the channel if :meth:`._get_redis_queue`
               returns a non-null result given ``args`` and ``kwargs``;
            3. It forward the frames read from the Redis channel(s) to the HTTP client;
            4. It periodically invokes :meth:`._should_stop` to cleanly
               terminate when the HTTP client socket is closed.

        """
        redis_queue = self._get_redis_queue(  # pylint: disable=assignment-from-none
            *args, **kwargs
        )

        if redis_queue:
            self.subscribe(redis_queue)

        try:
            for msg in self.listen():
                if self._should_stop():
                    break

                if redis_queue and msg.channel != redis_queue:
                    continue

                frame = msg.data
                if frame:
                    self.write(frame)
                    self.flush()
        finally:
            if redis_queue:
                self.unsubscribe(redis_queue)

    def _should_stop(self):
        """
        Utility method used by :meth:`._forward_stream` to automatically
        terminate when the client connection is closed (it can be overridden by
        the subclasses).
        """
        if self._finished:
            return True

        if self.request.connection and getattr(self.request.connection, 'stream', None):
            return self.request.connection.stream.closed()  # type: ignore

        return True

    def _on_error(self, e: Exception):
        """
        Handler for uncaught exceptions.
        """
        if isinstance(e, FileNotFoundError):
            self.set_status(404, str(e))
        elif isinstance(e, NotImplementedError):
            self.set_status(422, str(e))
        else:
            logger().exception(e)
            if isinstance(e, AssertionError):
                self.set_status(400, str(e))
            else:
                self.set_status(500, str(e))

        self.finish(json.dumps({"error": str(e)}))



from abc import ABC
from typing import Optional

from platypush.common.sensors import SensorDataType
from platypush.message.event import Event


class SensorDataEvent(Event, ABC):
    """
    Sensor events base class.
    """

    def __init__(
        self, *args, data: SensorDataType, source: Optional[str] = None, **kwargs
    ):
        """
        :param data: Sensor data.
        :param source: Sensor source - usually the plugin qualified name.
        """
        super().__init__(data=data, source=source, *args, **kwargs)


class SensorDataChangeEvent(SensorDataEvent):
    """
    Event triggered when a sensor has new data
    """


class SensorDataAboveThresholdEvent(SensorDataEvent):
    """
    Event triggered when a sensor's read goes above a configured threshold
    """


class SensorDataBelowThresholdEvent(SensorDataEvent):
    """
    Event triggered when a sensor's read goes below a configured threshold
    """


# vim:sw=4:ts=4:et:



from typing import List

from platypush.message.event import Event


class QrcodeScannedEvent(Event):
    """
    Event triggered when a QR-code or bar code is scanned.
    """

    def __init__(self, results: List[dict], *args, **kwargs):
        """
        :param results: List of decoded QR code results:

            .. schema:: qrcode.QrcodeDecodedResultSchema(many=True)
        """
        super().__init__(*args, results=results, **kwargs)


# vim:sw=4:ts=4:et:



from sqlalchemy import (
    Column,
    ForeignKey,
    Integer,
    JSON,
    String,
)

from platypush.common.db import is_defined

from .devices import Device


if not is_defined('cloud_instance'):

    class CloudInstance(Device):
        """
        Entity that maps a cloud node - like a Linode or AWS instance.
        """

        __tablename__ = 'cloud_instance'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )

        status = Column(String)
        instance_type = Column(String)
        ipv4_addresses = Column(JSON)
        ipv6_address = Column(String)
        group = Column(String)
        tags = Column(JSON)
        image = Column(String)
        region = Column(String)
        hypervisor = Column(String)
        uuid = Column(String)
        specs = Column(JSON)
        alerts = Column(JSON)
        backups = Column(JSON)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from sqlalchemy import Column, Integer, ForeignKey

from platypush.common.db import is_defined

from .sensors import NumericSensor


if not is_defined('humidity_sensor'):

    class HumiditySensor(NumericSensor):
        """
        A sensor that measures humidity.
        """

        __tablename__ = 'humidity_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('dew_point_sensor'):

    class DewPointSensor(NumericSensor):
        """
        A sensor that measures the dew point.
        """

        __tablename__ = 'dew_point_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



import functools
import inspect
import json
import os
import queue
import re
import subprocess
import tempfile
import threading
from abc import ABC, abstractmethod
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import requests

from platypush.config import Config
from platypush.context import get_plugin, get_backend
from platypush.message.event.media import MediaEvent
from platypush.plugins import RunnablePlugin, action
from platypush.utils import get_default_downloads_dir, get_plugin_name_by_class

from ._download import (
    DownloadState,
    DownloadThread,
    FileDownloadThread,
    YouTubeDownloadThread,
)
from ._resource import MediaResource
from ._state import PlayerState


class MediaPlugin(RunnablePlugin, ABC):
    """
    Generic plugin to interact with a media player.

    To start the local media stream service over HTTP you will also need the
    :class:`platypush.backend.http.HttpBackend` backend enabled.
    """

    # A media plugin can either be local or remote (e.g. control media on
    # another device)
    _is_local = True
    _NOT_IMPLEMENTED_ERR = NotImplementedError(
        'This method must be implemented in a derived class'
    )

    # Supported audio extensions
    audio_extensions = {
        '3gp',
        'aa',
        'aac',
        'aax',
        'act',
        'aiff',
        'amr',
        'ape',
        'au',
        'awb',
        'dct',
        'dss',
        'dvf',
        'flac',
        'gsm',
        'iklax',
        'ivs',
        'm4a',
        'm4b',
        'm4p',
        'mmf',
        'mp3',
        'mpc',
        'msv',
        'nmf',
        'nsf',
        'ogg,',
        'opus',
        'ra,',
        'raw',
        'sln',
        'tta',
        'vox',
        'wav',
        'wma',
        'wv',
        'webm',
        '8svx',
    }

    # Supported video extensions
    video_extensions = {
        'webm',
        'mkv',
        'flv',
        'vob',
        'ogv',
        'ogg',
        'drc',
        'gif',
        'gifv',
        'mng',
        'avi',
        'mts',
        'm2ts',
        'mov',
        'qt',
        'wmv',
        'yuv',
        'rm',
        'rmvb',
        'asf',
        'amv',
        'mp4',
        'm4p',
        'm4v',
        'mpg',
        'mp2',
        'mpeg',
        'mpe',
        'mpv',
        'm2v',
        'svi',
        '3gp',
        '3g2',
        'mxf',
        'roq',
        'nsv',
        'f4v',
        'f4p',
        'f4a',
        'f4b',
    }

    supported_media_plugins = [
        'media.mplayer',
        'media.omxplayer',
        'media.mpv',
        'media.vlc',
        'media.chromecast',
        'media.gstreamer',
    ]

    _supported_media_types = ['file', 'jellyfin', 'plex', 'torrent', 'youtube']
    _default_search_timeout = 60  # 60 seconds

    def __init__(
        self,
        media_dirs: Optional[List[str]] = None,
        download_dir: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        volume: Optional[Union[float, int]] = None,
        torrent_plugin: str = 'torrent',
        youtube_format: Optional[str] = 'bv[height<=?1080][ext=mp4]+ba',
        youtube_dl: str = 'yt-dlp',
        **kwargs,
    ):
        """
        :param media_dirs: Directories that will be scanned for media files when
            a search is performed (default: none)

        :param download_dir: Directory where external resources/torrents will be
            downloaded (default: ~/Downloads)

        :param env: Environment variables key-values to pass to the
            player executable (e.g. DISPLAY, XDG_VTNR, PULSE_SINK etc.)

        :param volume: Default volume for the player (default: None, maximum volume).

        :param torrent_plugin: Optional plugin to be used for torrent download. Possible values:

            - ``torrent`` - native ``libtorrent``-based plugin (default, recommended)
            - ``rtorrent`` - torrent support over rtorrent RPC/XML interface
            - ``webtorrent`` - torrent support over webtorrent (unstable)

        :param youtube_format: Select the preferred video/audio format for
            YouTube videos - and any media supported by youtube-dl or the
            selected fork. See the `youtube-dl documentation
            <https://github.com/ytdl-org/youtube-dl#format-selection>`_ for more
            info on supported formats. Default:
            ``bv*[height<=?1080][ext=mp4]+bestaudio/best`` - select the best
            mp4 video with a resolution <= 1080p, and the best audio format.

        :param youtube_dl: Path to the ``youtube-dl`` executable, used to
            extract information from YouTube videos and other media platforms.
            Default: ``yt-dlp``. The default has changed from ``youtube-dl`` to
            the ``yt-dlp`` fork because the former is badly maintained and its
            latest release was pushed in 2021.
        """

        super().__init__(**kwargs)

        if media_dirs is None:
            media_dirs = []
        player = None
        player_config = {}
        self._download_threads: Dict[Tuple[str, str], DownloadThread] = {}

        if self.__class__.__name__ == 'MediaPlugin':
            # Abstract class, initialize with the default configured player
            for plugin_name in Config.get_plugins().keys():
                if plugin_name in self.supported_media_plugins:
                    player = get_plugin(plugin_name)
                    if player and player.is_local():
                        # Local players have priority as default if configured
                        break
        else:
            player = self  # Derived concrete class

        if not player:
            raise AttributeError('No media plugin configured')

        media_dirs = media_dirs or player_config.get('media_dirs', [])

        if self.__class__.__name__ == 'MediaPlugin':
            # Populate this plugin with the actions of the configured player
            for act in player.registered_actions:
                setattr(self, act, getattr(player, act))
                self.registered_actions.add(act)

        self._env = env or {}
        self.media_dirs = set(
            filter(
                os.path.isdir,
                [os.path.abspath(os.path.expanduser(d)) for d in media_dirs],
            )
        )

        self.download_dir = os.path.abspath(
            os.path.expanduser(
                download_dir
                or player_config.get('download_dir')
                or get_default_downloads_dir()
            )
        )

        os.makedirs(self.download_dir, exist_ok=True)
        self._ytdl = youtube_dl
        self.media_dirs.add(self.download_dir)
        self.volume = volume
        self._videos_queue = []
        self._youtube_proc = None
        self.torrent_plugin = torrent_plugin
        self.youtube_format = youtube_format
        self._latest_resource: Optional[MediaResource] = None

    @staticmethod
    def _torrent_event_handler(evt_queue):
        def handler(event):
            # More than 5% of the torrent has been downloaded
            if event.args.get('progress', 0) > 5 and event.args.get('files'):
                evt_queue.put(event.args['files'])

        return handler

    def get_extractors(self):
        try:
            from yt_dlp.extractor import _extractors  # type: ignore
        except ImportError:
            self.logger.debug('yt_dlp not installed')
            return

        for _, obj_type in inspect.getmembers(_extractors):
            if (
                inspect.isclass(obj_type)
                and isinstance(getattr(obj_type, "_VALID_URL", None), str)
                and obj_type.__name__ != "GenericIE"
            ):
                yield obj_type

    def _is_youtube_resource(self, resource: str):
        return any(
            re.search(getattr(extractor, '_VALID_URL', '^$'), resource)
            for extractor in self.get_extractors()
        )

    def _get_youtube_best_thumbnail(self, info: Dict[str, dict]):
        thumbnails = info.get('thumbnails', {})
        if not thumbnails:
            return None

        # Preferred resolution
        for res in ((640, 480), (480, 360), (320, 240)):
            thumb = next(
                (
                    thumb
                    for thumb in thumbnails
                    if thumb.get('width') == res[0] and thumb.get('height') == res[1]
                ),
                None,
            )

            if thumb:
                return thumb.get('url')

        # Default fallback (best quality)
        return info.get('thumbnail')

    def _get_resource(self, resource: str):
        """
        :param resource: Resource to play/parse. Supported types:

            * Local files (format: ``file://<path>/<file>``)
            * Remote videos (format: ``https://<url>/<resource>``)
            * Torrents (format: Magnet links, Torrent URLs or local Torrent files)
            * Any URL that is supported by a yt_dlp extractor

        """

        if resource.startswith('file://'):
            path = resource[len('file://') :]
            assert os.path.isfile(path), f'File {path} not found'
            self._latest_resource = MediaResource(
                resource=resource,
                url=resource,
                title=os.path.basename(resource),
                filename=os.path.basename(resource),
            )
        elif self._is_youtube_resource(resource):
            info = self._get_youtube_info(resource)
            if info:
                url = info.get('url')
                if url:
                    resource = url
                    self._latest_resource = MediaResource(
                        resource=resource,
                        url=resource,
                        title=info.get('title'),
                        description=info.get('description'),
                        filename=info.get('filename'),
                        image=info.get('thumbnail'),
                        duration=float(info.get('duration') or 0) or None,
                        channel=info.get('channel'),
                        channel_url=info.get('channel_url'),
                        resolution=info.get('resolution'),
                        type=info.get('extractor'),
                    )
        elif resource.startswith('magnet:?'):
            self.logger.info(
                'Downloading torrent %s to %s', resource, self.download_dir
            )
            torrents = get_plugin(self.torrent_plugin)
            assert torrents, f'{self.torrent_plugin} plugin not configured'

            evt_queue = queue.Queue()
            torrents.download(
                resource,
                download_dir=self.download_dir,
                _async=True,
                is_media=True,
                event_hndl=self._torrent_event_handler(evt_queue),
            )

            resources = [f for f in evt_queue.get()]  # noqa: C416,R1721

            if resources:
                self._videos_queue = sorted(resources)
                resource = self._videos_queue.pop(0)
            else:
                raise RuntimeError(f'No media file found in torrent {resource}')

        assert resource, 'Unable to find any compatible media resource'
        return resource

    def _stop_torrent(self):
        try:
            torrents = get_plugin(self.torrent_plugin)
            assert torrents, f'{self.torrent_plugin} plugin not configured'
            torrents.quit()
        except Exception as e:
            self.logger.warning('Could not stop torrent plugin: %s', e)

    @action
    @abstractmethod
    def play(self, resource, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def pause(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def stop(self, *args, **kwargs):  # type: ignore
        super().stop()

    @action
    @abstractmethod
    def quit(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def voldown(self, step: Optional[float] = 5.0, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def volup(self, step: Optional[float] = 5.0, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def back(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def forward(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def next(self):
        """Play the next item in the queue"""
        self.stop()

        if self._videos_queue:
            video = self._videos_queue.pop(0)
            return self.play(video)

        return None

    @action
    @abstractmethod
    def toggle_subtitles(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def set_subtitles(self, filename, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def remove_subtitles(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def is_playing(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def load(self, resource, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def mute(self, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def seek(self, position, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def set_position(self, position, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def set_volume(self, volume):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    @abstractmethod
    def status(self):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def search(
        self,
        query: str,
        types: Optional[Iterable[str]] = None,
        queue_results: bool = False,
        autoplay: bool = False,
        timeout: float = _default_search_timeout,
    ):
        """
        Perform a video search.

        :param query: Query string, video name or partial name
        :param types: Video types to search (default: ``["youtube", "file", "torrent"]``)
        :param queue_results: Append the results to the current playing queue (default: False)
        :param autoplay: Play the first result of the search (default: False)
        :param timeout: Search timeout (default: 60 seconds)
        """

        results = {}
        results_queues = {}
        worker_threads = {}

        if types is None:
            types = self._supported_media_types

        for media_type in types:
            results[media_type] = []
            results_queues[media_type] = queue.Queue()
            search_hndl = self._get_search_handler_by_type(media_type)
            worker_threads[media_type] = threading.Thread(
                target=self._search_worker(
                    query=query,
                    search_hndl=search_hndl,
                    results_queue=results_queues[media_type],
                )
            )
            worker_threads[media_type].start()

        for media_type in types:
            try:
                items = results_queues[media_type].get(timeout=timeout)
                if isinstance(items, Exception):
                    raise items

                results[media_type].extend(items)
            except queue.Empty:
                self.logger.warning(
                    'Search for "%s" media type %s timed out', query, media_type
                )
            except Exception as e:
                self.logger.warning(
                    'Error while searching for "%s", media type %s', query, media_type
                )
                self.logger.exception(e)

        results = [
            {**result, 'type': media_type}
            for media_type in self._supported_media_types
            for result in results.get(media_type, [])
            if media_type in results
        ]

        if results:
            if queue_results:
                self._videos_queue = [_['url'] for _ in results]
                if autoplay:
                    self.play(self._videos_queue.pop(0))
            elif autoplay:
                self.play(results[0]['url'])

        return results

    @staticmethod
    def _search_worker(query, search_hndl, results_queue):
        def thread():
            try:
                results_queue.put(search_hndl.search(query))
            except Exception as e:
                results_queue.put(e)

        return thread

    def _get_search_handler_by_type(self, search_type: str):
        if search_type == 'file':
            from .search import LocalMediaSearcher

            return LocalMediaSearcher(self.media_dirs, media_plugin=self)
        if search_type == 'torrent':
            from .search import TorrentMediaSearcher

            return TorrentMediaSearcher(media_plugin=self)
        if search_type == 'youtube':
            from .search import YoutubeMediaSearcher

            return YoutubeMediaSearcher(media_plugin=self)
        if search_type == 'plex':
            from .search import PlexMediaSearcher

            return PlexMediaSearcher(media_plugin=self)
        if search_type == 'jellyfin':
            from .search import JellyfinMediaSearcher

            return JellyfinMediaSearcher(media_plugin=self)

        self.logger.warning('Unsupported search type: %s', search_type)
        return None

    @classmethod
    def is_video_file(cls, filename: str):
        return filename.lower().split('.')[-1] in cls.video_extensions

    @classmethod
    def is_audio_file(cls, filename: str):
        return filename.lower().split('.')[-1] in cls.audio_extensions

    def _get_info(self, resource: str):
        if self._is_youtube_resource(resource):
            return self.get_youtube_info(resource)

        return {'url': resource}

    @action
    def start_streaming(
        self, media: str, subtitles: Optional[str] = None, download: bool = False
    ):
        """
        Starts streaming local media over the specified HTTP port.
        The stream will be available to HTTP clients on
        `http://{this-ip}:{http_backend_port}/media/<media_id>`

        :param media: Media to stream
        :param subtitles: Path or URL to the subtitles track to be used
        :param download: Set to True if you prefer to download the file from
            the streaming link instead of streaming it
        :return: dict containing the streaming URL.Example:

        .. code-block:: json

            {
                "id": "0123456abcdef.mp4",
                "source": "file:///mnt/media/movies/movie.mp4",
                "mime_type": "video/mp4",
                "url": "http://192.168.1.2:8008/media/0123456abcdef.mp4"
            }

        """
        return self._start_streaming(media, subtitles=subtitles, download=download)

    def _start_streaming(
        self, media: str, subtitles: Optional[str] = None, download: bool = False
    ):
        http = get_backend('http')
        assert http, f'Unable to stream {media}: HTTP backend not configured'

        self.logger.info('Starting streaming %s', media)
        response = requests.put(
            f'{http.local_base_url}/media' + ('?download' if download else ''),
            json={'source': media, 'subtitles': subtitles},
            timeout=300,
        )

        assert response.ok, response.text or response.reason
        return response.json()

    @action
    def stop_streaming(self, media_id: str):
        http = get_backend('http')
        assert http, f'Unable to stop streaming {media_id}: HTTP backend not configured'

        response = requests.delete(
            f'{http.local_base_url}/media/{media_id}', timeout=30
        )

        assert response.ok, response.text or response.reason
        return response.json()

    def _get_youtube_info(self, url, youtube_format: Optional[str] = None):
        ytdl_cmd = [
            self._ytdl,
            *(
                ['-f', youtube_format or self.youtube_format]
                if youtube_format or self.youtube_format
                else []
            ),
            '-j',
            '-g',
            url,
        ]

        self.logger.info('Executing command %s', ' '.join(ytdl_cmd))
        with subprocess.Popen(ytdl_cmd, stdout=subprocess.PIPE) as ytdl:
            output = ytdl.communicate()[0].decode().strip()
            ytdl.wait()

        self.logger.debug('yt-dlp output: %s', output)
        lines = output.split('\n')

        if not lines:
            self.logger.warning('No output from yt-dlp')
            return None

        stream_url = lines[1] if len(lines) > 2 else lines[0]
        info = lines[-1]
        return {
            **json.loads(info),
            'url': stream_url,
        }

    @staticmethod
    def get_youtube_id(url: str) -> Optional[str]:
        patterns = [
            re.compile(pattern)
            for pattern in [
                r'https?://www.youtube.com/watch\?v=([^&#]+)',
                r'https?://youtube.com/watch\?v=([^&#]+)',
                r'https?://youtu.be/([^&#/]+)',
                r'youtube:video:([^&#:])',
            ]
        ]

        for pattern in patterns:
            m = pattern.search(url)
            if m:
                return m.group(1)

        return None

    @action
    def get_youtube_info(self, url):
        # Legacy conversion for Mopidy YouTube URIs
        m = re.match('youtube:video:(.*)', url)
        if m:
            url = f'https://www.youtube.com/watch?v={m.group(1)}'

        with subprocess.Popen([self._ytdl, '-j', url], stdout=subprocess.PIPE) as proc:
            if proc.stdout is None:
                return None

            return proc.stdout.read().decode("utf-8", "strict")[:-1]

    @action
    def get_info(self, resource: str):
        return self._get_info(resource)

    @action
    def get_media_file_duration(self, filename):
        """
        Get the duration of a media file in seconds. Requires ffmpeg
        """

        if filename.startswith('file://'):
            filename = filename[7:]

        with subprocess.Popen(
            ["ffprobe", filename], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        ) as result:
            if not result.stdout:
                return 0

            return functools.reduce(
                lambda t, t_i: t + t_i,
                [
                    float(t) * pow(60, i)
                    for (i, t) in enumerate(
                        re.search(
                            r'^Duration:\s*([^,]+)',
                            [
                                x.decode()
                                for x in result.stdout.readlines()
                                if "Duration" in x.decode()
                            ]
                            .pop()
                            .strip(),
                        )
                        .group(1)  # type: ignore
                        .split(':')[::-1]
                    )
                ],
            )

    @action
    def download(
        self,
        url: str,
        filename: Optional[str] = None,
        directory: Optional[str] = None,
        timeout: int = 10,
        sync: bool = False,
        only_audio: bool = False,
        youtube_format: Optional[str] = None,
    ):
        """
        Download a media URL to a local file on the Platypush host (yt-dlp
        required for YouTube URLs).

        This action is non-blocking and returns the path to the downloaded file
        once the download is initiated.

        You can then subscribe to these events to monitor the download progress:

            - :class:`platypush.message.event.media.MediaDownloadStartedEvent`
            - :class:`platypush.message.event.media.MediaDownloadProgressEvent`
            - :class:`platypush.message.event.media.MediaDownloadErrorEvent`
            - :class:`platypush.message.event.media.MediaDownloadPausedEvent`
            - :class:`platypush.message.event.media.MediaDownloadResumedEvent`
            - :class:`platypush.message.event.media.MediaDownloadCancelledEvent`
            - :class:`platypush.message.event.media.MediaDownloadCompletedEvent`

        :param url: Media URL.
        :param filename: Media filename (default: inferred from the URL basename).
        :param directory: Destination directory (default: ``download_dir``).
        :param timeout: Network timeout in seconds (default: 10).
        :param sync: If set to True, the download will be synchronous and the
            action will return only when the download is completed.
        :param youtube_format: Override the default ``youtube_format`` setting.
        :return: The absolute path to the downloaded file.
        """
        path = self._get_download_path(
            url, directory=directory, filename=filename, youtube_format=youtube_format
        )

        if self._is_youtube_resource(url):
            dl_thread = self._download_youtube_url(
                url, path, youtube_format=youtube_format, only_audio=only_audio
            )
        else:
            if only_audio:
                self.logger.warning(
                    'Only audio download is not supported for non-YouTube URLs'
                )

            dl_thread = self._download_url(url, path, timeout=timeout)

        if sync:
            dl_thread.join()

        return path

    @action
    def pause_download(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Pause a download in progress.

        Either the URL or the path must be specified.

        :param url: URL of the download.
        :param path: Path of the download (default: any path associated with the URL).
        """
        for thread in self._get_downloads(url=url, path=path):
            thread.pause()

    @action
    def resume_download(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Resume a paused download.

        Either the URL or the path must be specified.

        :param url: URL of the download.
        :param path: Path of the download (default: any path associated with the URL).
        """
        for thread in self._get_downloads(url=url, path=path):
            thread.resume()

    @action
    def cancel_download(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Cancel a download in progress.

        Either the URL or the path must be specified.

        :param url: URL of the download.
        :param path: Path of the download (default: any path associated with the URL).
        """
        for thread in self._get_downloads(url=url, path=path):
            thread.stop()

    @action
    def clear_downloads(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Clear completed/cancelled downloads from the queue.

        :param url: URL of the download (default: all downloads).
        :param path: Path of the download (default: any path associated with the URL).
        """
        threads = (
            self._get_downloads(url=url, path=path)
            if url
            else list(self._download_threads.values())
        )

        for thread in threads:
            if thread.state not in (DownloadState.COMPLETED, DownloadState.CANCELLED):
                continue

            dl = self._download_threads.pop((thread.url, thread.path), None)
            if dl:
                dl.clear()

    @action
    def get_downloads(self, url: Optional[str] = None, path: Optional[str] = None):
        """
        Get the download threads.

        :param url: URL of the download (default: all downloads).
        :param path: Path of the download (default: any path associated with the URL).
        :return: .. schema:: media.download.MediaDownloadSchema(many=True)
        """
        from platypush.schemas.media.download import MediaDownloadSchema

        return MediaDownloadSchema().dump(
            (
                self._get_downloads(url=url, path=path)
                if url
                else list(self._download_threads.values())
            ),
            many=True,
        )

    def _get_downloads(self, url: Optional[str] = None, path: Optional[str] = None):
        assert url or path, 'URL or path must be specified'
        threads = []

        if url and path:
            path = os.path.expanduser(path)
            thread = self._download_threads.get((url, path))
            if thread:
                threads = [thread]
        elif url:
            threads = [
                thread
                for (url_, _), thread in self._download_threads.items()
                if url_ == url
            ]
        elif path:
            path = os.path.expanduser(path)
            threads = [
                thread
                for (_, path_), thread in self._download_threads.items()
                if path_ == path
            ]

        assert threads, f'No matching downloads found for [url={url}, path={path}]'
        return threads

    def _get_download_path(
        self,
        url: str,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        youtube_format: Optional[str] = None,
    ) -> str:
        if not directory:
            directory = self.download_dir

        directory = os.path.expanduser(directory)
        youtube_format = youtube_format or self.youtube_format

        if self._is_youtube_resource(url):
            with subprocess.Popen(
                [
                    self._ytdl,
                    *(
                        [
                            '-f',
                            youtube_format,
                        ]
                        if youtube_format
                        else []
                    ),
                    '-O',
                    '%(title)s.%(ext)s',
                    url,
                ],
                stdout=subprocess.PIPE,
            ) as proc:
                assert proc.stdout, 'yt-dlp stdout is None'
                filename = proc.stdout.read().decode()[:-1]

        if not filename:
            filename = url.split('/')[-1]

        return os.path.join(directory, filename)

    def _download_url(self, url: str, path: str, timeout: int) -> FileDownloadThread:
        download_thread = FileDownloadThread(
            url=url,
            path=path,
            timeout=timeout,
            on_start=self._on_download_start,
            post_event=self._post_event,
            stop_event=self._should_stop,
        )

        self._start_download(download_thread)
        return download_thread

    def _download_youtube_url(
        self,
        url: str,
        path: str,
        youtube_format: Optional[str] = None,
        only_audio: bool = False,
    ) -> YouTubeDownloadThread:
        download_thread = YouTubeDownloadThread(
            url=url,
            path=path,
            ytdl=self._ytdl,
            only_audio=only_audio,
            youtube_format=youtube_format or self.youtube_format,
            on_start=self._on_download_start,
            post_event=self._post_event,
            stop_event=self._should_stop,
        )

        self._start_download(download_thread)
        return download_thread

    def _on_download_start(self, thread: DownloadThread):
        self._download_threads[thread.url, thread.path] = thread

    def _start_download(self, thread: DownloadThread):
        if (thread.url, thread.path) in self._download_threads:
            self.logger.warning(
                'A download of %s to %s is already in progress', thread.url, thread.path
            )
            return

        thread.start()

    def _post_event(self, event_type: Type[MediaEvent], **kwargs):
        evt = event_type(
            player=get_plugin_name_by_class(self.__class__), plugin=self, **kwargs
        )
        self._bus.post(evt)

    def is_local(self):
        return self._is_local

    @staticmethod
    def get_subtitles_file(subtitles: Optional[str] = None):
        if not subtitles:
            return None

        if subtitles.startswith('file://'):
            subtitles = subtitles[len('file://') :]
        if os.path.isfile(subtitles):
            return os.path.abspath(subtitles)

        content = requests.get(subtitles, timeout=20).content
        f = tempfile.NamedTemporaryFile(
            prefix='media_subs_', suffix='.srt', delete=False
        )

        with f:
            f.write(content)
        return f.name

    def main(self):
        self.wait_stop()


__all__ = [
    'DownloadState',
    'MediaPlugin',
    'PlayerState',
]


# vim:sw=4:ts=4:et:



from typing import Optional

from datetime import datetime
from dateutil.tz import tzutc
from marshmallow import fields, pre_dump
from marshmallow.schema import Schema
from marshmallow.validate import OneOf

from platypush.schemas import DateTime


def _get_precip_type(response: dict) -> Optional[str]:
    if response.get('snow', {}).get('1h', 0) > 0:
        return 'snow'
    if response.get('rain', {}).get('1h', 0) > 0:
        return 'rain'
    return None


class WeatherSchema(Schema):
    """
    Schema for weather data.
    """

    time = DateTime(
        required=True,
        attribute='dt',
        metadata={
            'description': 'Time of the weather condition',
            'example': '2020-01-01T12:00:00+00:00',
        },
    )

    summary = fields.Function(
        lambda obj: obj.get('weather', [{'main': 'Unknown'}])[0].get('main', 'Unknown'),
        metadata={
            'description': 'Summary of the weather condition',
            'example': 'Cloudy',
        },
    )

    icon = fields.Function(
        lambda obj: obj.get('weather', [{'icon': 'unknown'}])[0].get('icon', 'unknown'),
        metadata={
            'description': 'Icon representing the weather condition',
            'example': 'cloudy',
        },
    )

    precip_intensity = fields.Function(
        lambda obj: obj.get('rain', obj.get('snow', {})).get('1h', 0),
        metadata={
            'description': 'Intensity of the precipitation',
            'example': 0.0,
        },
    )

    precip_type = fields.Function(
        _get_precip_type,
        metadata={
            'description': 'Type of precipitation',
            'example': 'rain',
        },
    )

    temperature = fields.Function(
        lambda obj: obj.get('main', {}).get('temp'),
        metadata={
            'description': 'Temperature in the configured unit of measure',
            'example': 10.0,
        },
    )

    apparent_temperature = fields.Function(
        lambda obj: obj.get('main', {}).get('feels_like'),
        metadata={
            'description': 'Apparent temperature in the configured unit of measure',
            'example': 9.0,
        },
    )

    humidity = fields.Function(
        lambda obj: obj.get('main', {}).get('humidity'),
        metadata={
            'description': 'Humidity percentage, between 0 and 100',
            'example': 30,
        },
    )

    pressure = fields.Function(
        lambda obj: obj.get('main', {}).get('pressure'),
        metadata={
            'description': 'Pressure in hPa',
            'example': 1000.0,
        },
    )

    wind_speed = fields.Function(
        lambda obj: obj.get('wind', {}).get('speed'),
        metadata={
            'description': 'Wind speed in the configured unit of measure',
            'example': 10.0,
        },
    )

    wind_direction = fields.Function(
        lambda obj: obj.get('wind', {}).get('deg'),
        metadata={
            'description': 'Wind direction in degrees',
            'example': 180.0,
        },
    )

    wind_gust = fields.Function(
        lambda obj: obj.get('wind', {}).get('gust'),
        metadata={
            'description': 'Wind gust in the configured unit of measure',
            'example': 15.0,
        },
    )

    cloud_cover = fields.Function(
        lambda obj: obj.get('clouds', {}).get('all'),
        metadata={
            'description': 'Cloud cover percentage between 0 and 100',
            'example': 0.5,
        },
    )

    visibility = fields.Float(
        metadata={
            'description': 'Visibility in meters',
            'example': 2000.0,
        },
    )

    sunrise = DateTime(
        metadata={
            'description': 'Sunrise time',
            'example': '2020-01-01T06:00:00+00:00',
        },
    )

    sunset = DateTime(
        metadata={
            'description': 'Sunset time',
            'example': '2020-01-01T18:00:00+00:00',
        },
    )

    units = fields.String(
        missing='metric',
        validate=OneOf(['metric', 'imperial']),
        metadata={
            'description': 'Unit of measure',
            'example': 'metric',
        },
    )

    @staticmethod
    def _timestamp_to_dt(timestamp: float) -> datetime:
        return datetime.fromtimestamp(timestamp, tz=tzutc())

    @pre_dump
    def _pre_dump(self, data: dict, **_) -> dict:
        sun_data = data.pop('sys', {})
        sunrise = sun_data.get('sunrise')
        sunset = sun_data.get('sunset')

        if sunrise is not None:
            data['sunrise'] = self._timestamp_to_dt(sunrise)
        if sunset is not None:
            data['sunset'] = self._timestamp_to_dt(sunset)

        return data


class WeatherReportSchema(Schema):
    """
    Schema for full weather reports.
    """

    current = fields.Nested(WeatherSchema, required=True)
    forecast = fields.List(fields.Nested(WeatherSchema), required=True)



from marshmallow import EXCLUDE, fields, pre_load
from marshmallow.schema import Schema

from platypush.schemas import DateTime


class GpsDeviceSchema(Schema):
    """
    Schema for a GPS device.
    """

    # pylint: disable=too-few-public-methods
    class Meta:  # type: ignore
        """
        Exclude unknown fields from the deserialized output.
        """

        unknown = EXCLUDE

    path = fields.String(
        required=True,
        metadata={
            "description": "Device path",
            "example": "/dev/ttyUSB0",
        },
    )

    activated = DateTime(
        metadata={
            "description": "Device activation status",
            "example": True,
        },
    )

    native = fields.Boolean(
        metadata={
            "description": "Device native status",
            "example": False,
        },
    )

    baudrate = fields.Integer(
        data_key="bps",
        metadata={
            "description": "Device baudrate",
            "example": 9600,
        },
    )

    parity = fields.String(
        metadata={
            "description": "Device parity",
            "example": "N",
        },
    )

    stopbits = fields.Integer(
        metadata={
            "description": "Device stopbits",
            "example": 1,
        },
    )

    cycle = fields.Integer(
        metadata={
            "description": "Device cycle",
            "example": 1,
        },
    )

    driver = fields.String(
        metadata={
            "description": "Device driver",
            "example": "NMEA",
        },
    )

    subtype = fields.String(
        metadata={
            "description": "Device subtype",
            "example": "AXN_2.31_3339_13101700,5632,PA6H,1.0",
        },
    )

    mode = fields.String(
        validate=lambda mode: mode in ["NO_FIX", "TWO_D", "THREE_D"],
        metadata={
            "description": "Device mode, one of NO_FIX, TWO_D, THREE_D",
            "example": "3D",
        },
    )

    @pre_load
    def pre_load(self, data, **_):
        from platypush.plugins.gps import DeviceMode

        if data and data.get("mode"):
            data["mode"] = DeviceMode(data["mode"]).value
        return data


class GpsStatusSchema(Schema):
    """
    Schema for the GPS status.
    """

    latitude = fields.Float(
        metadata={
            "description": "Latitude",
            "example": 45.4642,
        },
    )

    longitude = fields.Float(
        metadata={
            "description": "Longitude",
            "example": 9.1900,
        },
    )

    altitude = fields.Float(
        metadata={
            "description": "Altitude (in meters)",
            "example": 100,
        },
    )

    speed = fields.Float(
        metadata={
            "description": "Measured speed, if available (in km/h)",
            "example": 10,
        },
    )

    satellites_used = fields.Integer(
        metadata={
            "description": "Number of satellites used for the fix",
            "example": 4,
        },
    )

    timestamp = DateTime(
        metadata={
            "description": "Timestamp of the last GPS update",
            "example": "2021-08-01T00:00:00",
        },
    )

    devices = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(GpsDeviceSchema),
        metadata={
            "description": "Available GPS devices",
            "example": {
                "/dev/ttyUSB0": {
                    "path": "/dev/ttyUSB0",
                    "activated": "2021-08-01T00:00:00",
                    "native": False,
                    "baudrate": 9600,
                    "parity": "N",
                    "stopbits": 1,
                    "cycle": 1,
                    "driver": "NMEA",
                    "subtype": "AXN_2.31_3339_13101700,5632,PA6H,1.0",
                    "mode": "3D",
                }
            },
        },
    )



from enum import Enum
import json
from typing import Final, Optional, Union

import numpy as np

from ._base import SoundBase


class WaveShape(Enum):
    """
    Supported audio wave shapes.
    """

    SIN = 'sin'
    SQUARE = 'square'
    SAWTOOTH = 'sawtooth'
    TRIANG = 'triang'


class Sound(SoundBase):
    """
    Models a basic synthetic sound that can be played through an audio device
    """

    _DEFAULT_MID_A_FREQUENCY: Final[float] = 440.0

    def __init__(
        self,
        *args,
        midi_note: Optional[Union[str, int]] = None,
        frequency: Optional[float] = None,
        phase: float = 0,
        duration: Optional[float] = None,
        delay: float = 0,
        shape: WaveShape = WaveShape.SIN,
        **kwargs,
    ):
        """
        You can construct a sound either from a MIDI note or a base frequency,
        as well as the shape of the output wave.

        :param midi_note: MIDI note code, see `this chart
            <https://newt.phys.unsw.edu.au/jw/graphics/notes.GIF>`_.
        :param frequency: Sound base frequency in Hz
        :param phase: Wave phase shift as a multiple of pi (default: 0.0)
        :param duration: Note duration in seconds. Default: keep until
            release/pause/stop
        :param delay: Sound delay in seconds, calculated from the moment the
            command execution starts. Default: 0.
        :param shape: Wave shape. Possible values: "``sin``", "``square``",
            "``sawtooth``" or "``triang``" (see :class:`WaveShape`).
            Default: "``sin``"
        """

        super().__init__(*args, **kwargs)
        invalid_request = RuntimeError(
            'Please specify either a MIDI note or a base frequency'
        )

        if midi_note and frequency:
            raise invalid_request

        if midi_note:
            self.midi_note = self.get_midi_note(midi_note)
            self.frequency = self.note_to_freq(midi_note=midi_note)
        elif frequency:
            self.frequency = frequency
            self.midi_note = self.freq_to_note(frequency=frequency)
        else:
            raise invalid_request

        self.phase = phase
        self.duration = duration
        self.delay = delay
        self.shape = WaveShape(shape)

    def _get_left_audio_pad(
        self, sample_rate: float, t_start: float, t_end: float
    ) -> int:
        """
        Get the size of the audio wave left zero-pad given in function of its
        ``delay``, ``sample_rate``, ``t_start`` and ``t_end``.
        """
        return round(max(0, min(t_end, self.delay) - t_start) * sample_rate)

    def _get_right_audio_pad(
        self, sample_rate: float, t_start: float, t_end: float
    ) -> int:
        """
        Get the size of the audio wave right zero-pad given its declared
        ``delay`` in function of ``t_start`` and ``t_end``.
        """
        if not self.duration:
            return 0

        duration = self.delay + self.duration
        if t_end <= duration:
            return 0

        return round((t_end - max(t_start, duration)) * sample_rate)

    def _get_audio_pad(
        self, sample_rate: float, t_start: float, t_end: float
    ):  # -> Tuple[NDArray[np.floating], NDArray[np.floating]]:
        """
        Return the left and right audio pads for a given audio length as a
        ``(left, right)`` tuple of numpy zero-filled arrays.
        """
        return tuple(
            np.zeros([pad_size, 1])
            for pad_size in (
                self._get_left_audio_pad(
                    sample_rate=sample_rate, t_start=t_start, t_end=t_end
                ),
                self._get_right_audio_pad(
                    sample_rate=sample_rate, t_start=t_start, t_end=t_end
                ),
            )
        )

    def _generate_wave(self, x):
        """
        Generate a raw audio wave as a numpy array of floating between -1 and 1
        given ``x`` as a set of timestamp samples.
        """
        if self.shape in (WaveShape.SIN, WaveShape.SQUARE):
            wave = np.sin((2 * np.pi * self.frequency * x) + np.pi * self.phase)

            if self.shape == WaveShape.SQUARE:
                wave[wave < 0] = -0.95
                wave[wave >= 0] = 0.95
        elif self.shape in (WaveShape.SAWTOOTH, WaveShape.TRIANG):
            wave = 2 * (self.frequency * x - np.floor(0.5 + self.frequency * x))
            if self.shape == WaveShape.TRIANG:
                wave = 2 * np.abs(wave) - 1
        else:
            raise RuntimeError(
                f'Unsupported wave shape: {self.shape}. '
                f'Supported values: {[s.value for s in WaveShape]}'
            )

        return wave

    def get_wave(
        self,
        sample_rate: float,
        t_start: float = 0,
        t_end: float = 0,
        **_,
    ):  # -> NDArray[np.floating]:
        """
        Get the wave binary data associated to this sound

        :param t_start: Start offset for the wave in seconds. Default: 0
        :param t_end: End offset for the wave in seconds. Default: 0
        :param sample_rate: Audio sample rate. Default: 44100 Hz
        :returns: A ``numpy.ndarray[(t_end-t_start)*sample_rate, 1]``
            with the raw float values
        """

        assert self.frequency is not None, 'The sound has no associated base frequency'
        if t_start > t_end:
            return np.array([])

        left_pad, right_pad = self._get_audio_pad(
            sample_rate=sample_rate, t_start=t_start, t_end=t_end
        )
        t_start = min(t_end, t_start + (left_pad.shape[0] / sample_rate))
        t_end = max(t_start, t_end - (right_pad.shape[0] / sample_rate))
        actual_n_samples = abs(round((t_end - t_start) * sample_rate))
        wave_length = max(t_start, self.delay - t_start)

        if self.duration is not None:
            wave_length = min(wave_length, self.duration - self.delay)

        x = np.linspace(
            max(t_start, self.delay - t_start),
            t_end,
            actual_n_samples,
        ).reshape(-1, 1)

        return self.gain * np.array(
            (
                *left_pad,
                *self._generate_wave(x),
                *right_pad,
            )
        )

    def __iter__(self):
        """
        Iterates over the sound's attributes and returns key-value pairs.
        """
        for attr in ['midi_note', 'frequency', 'volume', 'duration', 'ref_frequency']:
            yield attr, getattr(self, attr)

    def __str__(self):
        """
        :return: A JSON-string representation of the sound dictionary.
        """
        return json.dumps(dict(self))

    @classmethod
    def build(cls, *args, **kwargs) -> "Sound":
        """
        Construct a sound object either from a JSON representation or a
        key-value representation
        """

        if args:
            if isinstance(args[0], cls):
                return args[0]
            if isinstance(args[0], str):
                kwargs = json.loads(args[0])
            elif isinstance(args[0], dict):
                kwargs = args[0]

        if kwargs:
            return Sound(**kwargs)

        raise RuntimeError(f'Usage: {__doc__}')



from marshmallow import fields, EXCLUDE
from marshmallow.schema import Schema
from marshmallow.validate import Range


device_types = [
    'Hub',
    'Hub Plus',
    'Hub Mini',
    'Bot',
    'Curtain',
    'Plug',
    'Meter',
    'Humidifier',
    'Smart Fan',
    'Air Conditioner',
    'TV',
    'Light',
    'IPTV / Streamer',
    'Set Top Box',
    'DVD',
    'Fan',
    'Projector',
    'Camera',
    'Air Purifier',
    'Speaker',
    'Water Heater',
    'Vacuum Cleaner',
    'Remote',
    'Others',
]

remote_types = [
    'Air Conditioner',
    'TV',
    'Light',
    'IPTV / Streamer',
    'Set Top Box',
    'DVD',
    'Fan',
    'Projector',
    'Camera',
    'Air Purifier',
    'Speaker',
    'Water Heater',
    'Vacuum Cleaner',
    'Others',
]


class ColorField(fields.Field):
    """
    Utility field class for color values.
    """

    def _serialize(self, value: str, *_, **__):
        """
        Convert a hex native color value (``ff0000``) to the format exposed by
        the SwitchBot API (``255:0:0``).
        """
        if not value:
            return None
        # fmt: off
        return ''.join([f'{int(i):02x}' for i in value.split(':')])

    def _deserialize(self, value: str, *_, **__):
        """
        Convert a SwitchBot API color value (``255:0:0``) to the hex native
        format (``ff0000``).
        """
        if not value:
            return None

        value = value.lstrip('#')
        # fmt: off
        return ':'.join(
            [str(int(value[i:i+2], 16)) for i in range(0, len(value) - 1, 2)]
        )


class DeviceSchema(Schema):
    """
    Base class for SwitchBot device schemas.
    """

    class Meta:
        """
        Ignore unknown fields.
        """

        unknown = EXCLUDE

    id = fields.String(
        attribute='deviceId',
        required=True,
        metadata={'description': 'Device unique ID'},
    )
    name = fields.String(
        attribute='deviceName', metadata={'description': 'Device name'}
    )
    device_type = fields.String(
        attribute='deviceType',
        metadata={'description': f'Default types: [{", ".join(device_types)}]'},
    )
    remote_type = fields.String(
        attribute='remoteType',
        metadata={'description': f'Default types: [{", ".join(remote_types)}]'},
    )
    hub_id = fields.String(
        attribute='hubDeviceId',
        metadata={'description': 'Parent hub device unique ID'},
    )
    cloud_service_enabled = fields.Boolean(
        attribute='enableCloudService',
        metadata={
            'description': 'True if cloud access is enabled on this device,'
            'False otherwise. Only cloud-enabled devices can be '
            'controlled from the switchbot plugin.'
        },
    )
    is_calibrated = fields.Boolean(
        attribute='calibrate',
        metadata={
            'description': '[Curtain devices only] Set to True if the device '
            'has been calibrated, False otherwise'
        },
    )
    open_direction = fields.String(
        attribute='openDirection',
        metadata={
            'description': '[Curtain devices only] Direction where the curtains '
            'will be opened ("left" or "right")'
        },
    )
    is_virtual = fields.Boolean(
        metadata={
            'description': 'True if this is a virtual device, i.e. a device '
            'with an IR remote configuration but not managed directly by '
            'the Switchbot bridge'
        }
    )


class DeviceStatusSchema(DeviceSchema):
    """
    Schema for SwitchBot devices status.
    """

    on = fields.Boolean(
        attribute='power',
        metadata={'description': 'True if the device is on, False otherwise'},
    )
    voltage = fields.Float(
        allow_none=True,
        metadata={
            'description': '[Plug devices only] Voltage of the device, measured '
            'in volts'
        },
    )
    power = fields.Float(
        attribute='weight',
        allow_none=True,
        metadata={
            'description': '[Plug devices only] Consumed power, measured in watts'
        },
    )
    current = fields.Float(
        attribute='electricCurrent',
        allow_none=True,
        metadata={
            'description': '[Plug devices only] Device current at the moment, '
            'measured in amperes'
        },
    )
    active_time = fields.Int(
        attribute='electricityOfDay',
        allow_none=True,
        metadata={
            'description': '[Plug devices only] How long the device has been '
            'absorbing during a day, measured in minutes'
        },
    )
    moving = fields.Boolean(
        metadata={
            'description': '[Curtain devices only] True if the device is '
            'moving, False otherwise'
        }
    )
    position = fields.Int(
        attribute='slidePosition',
        allow_none=True,
        metadata={
            'description': '[Curtain devices only] Position of the device on '
            'the curtain rail, between 0% (open) and 100% (closed)'
        },
    )
    locked = fields.Boolean(
        attribute='lockState',
        metadata={'description': '[Lock devices only] True if the lock is on'},
    )
    door_open = fields.Boolean(
        attribute='doorState',
        metadata={
            'description': '[Lock devices only] True if the door is open, False otherwise'
        },
    )
    brightness = fields.Int(
        metadata={
            'description': '[Light devices only] Light brightness, between 1 and 100'
        },
        allow_none=True,
        validate=Range(min=1, max=100),
    )
    color = ColorField(
        allow_none=True,
        metadata={
            'description': '[Light devices only] Color, expressed as a hex string (e.g. FF0000)'
        },
    )
    color_temperature = fields.Int(
        attribute='colorTemperature',
        allow_none=True,
        validate=Range(min=2700, max=6500),
        metadata={
            'description': '[Light devices only] Color temperature, between 2700 and 6500'
        },
    )
    temperature = fields.Float(
        allow_none=True,
        metadata={
            'description': '[Meter/humidifier/Air conditioner devices only] '
            'Temperature in Celsius'
        },
    )
    humidity = fields.Float(
        allow_none=True,
        metadata={'description': '[Meter/humidifier devices only] Humidity in %'},
    )
    fan_speed = fields.Int(
        allow_none=True,
        metadata={'description': '[Air conditioner devices only] Speed of the fan'},
    )
    nebulization_efficiency = fields.Float(
        attribute='nebulizationEfficiency',
        allow_none=True,
        metadata={
            'description': '[Humidifier devices only] Nebulization efficiency in %'
        },
    )
    auto = fields.Boolean(
        metadata={'description': '[Humidifier devices only] True if auto mode is on'}
    )
    child_lock = fields.Boolean(
        attribute='childLock',
        metadata={'description': '[Humidifier devices only] True if safety lock is on'},
    )
    sound = fields.Boolean(
        metadata={'description': '[Humidifier devices only] True if sound is muted'}
    )
    low_water = fields.Boolean(
        attribute='lackWater',
        metadata={
            'description': '[Humidifier devices only] True if the device is low on water'
        },
    )
    mode = fields.Int(
        metadata={'description': '[Fan/Air conditioner devices only] Fan mode'}
    )
    speed = fields.Float(
        metadata={'description': '[Smart fan devices only] Fan speed, between 1 and 4'}
    )
    swinging = fields.Boolean(
        attribute='shaking',
        metadata={
            'description': '[Smart fan devices only] True if the device is swinging'
        },
    )
    swing_direction = fields.Int(
        attribute='shakeCenter',
        metadata={'description': '[Smart fan devices only] Swing direction'},
    )
    swing_range = fields.Float(
        attribute='shakeRange',
        metadata={
            'description': '[Smart fan devices only] Swing range angle, between 0 and 120'
        },
    )


class SceneSchema(Schema):
    """
    Schema for SwitchBot scenes.
    """

    id = fields.String(
        attribute='sceneId', required=True, metadata={'description': 'Scene ID'}
    )
    name = fields.String(attribute='sceneName', metadata={'description': 'Scene name'})
